/*******************************************************************************
 $Author: dmc $ 
 $Revision: 2477 $
 $Date: 2019-01-16 20:51:17 -0400 (Wed, 16 Jan 2019) $
 $HeadURL: http://support2.silvermountaintargets.com/svn/smt/smt.common.lib/mic_mapping.js $
*******************************************************************************/



// mic_mapping utilities
//
// shot.counters[] = array of readings from physical (electrical) channels, e.g.
//    pin "MIK1"..."MIK8"
// res.mic[] = array of readings from logical mic channels (remapped if necessary from electrical channels)
//
// phys_to_logical()

// mic mapping string of e.g. 87543612 means that
// electrical channel number 1 = MIK1 goes to logical channel 8
// ...
// electrical channel number 3 = MIK3 goes to logical channel 5
//
// if "0" is used, that throws away that electrical reading

function phys_to_logical(mic_mapping,phys_counters) {
	// apply mic mapping
	var mic = []; for (var i = 0; i < phys_counters.length; i++) mic[i] = -1;

	for (var i = 0; i < phys_counters.length; i++) {
		var dest = Number(mic_mapping.charAt(i));
		if (!isNaN(dest) && dest >= 1 && dest <= 8) mic[dest-1] = phys_counters[i];
	}
	return mic;
}
exports.phys_to_logical = phys_to_logical;

// logical_to_phys()
//
function logical_to_phys(mic_mapping,logical_counters) {
	// apply inverted mic mapping
	var phys_counters = [];
	for (var i = 0; i < logical_counters.length; i++) phys_counters[i] = -1;

	for (var i = 0; i < logical_counters.length; i++) {
		var dest = Number(mic_mapping.charAt(i));
		if (!isNaN(dest) && dest >= 1 && dest <= 8) phys_counters[i] = logical_counters[dest-1];
	}
	return phys_counters;
}
exports.logical_to_phys = logical_to_phys;

// logical_mics_in_use()
//

function logical_mics_in_use(mic_mapping) {
	// apply inverted mic mapping
	var mics_in_use = [false, false, false, false, false, false, false, false];

	for (var i = 0; i < mic_mapping.length; i++) {
		var dest = Number(mic_mapping.charAt(i));
		if (!isNaN(dest) && dest >= 1 && dest <= 8) mics_in_use[dest-1] = true;
	}
	return mics_in_use;
}
exports.logical_mics_in_use = logical_mics_in_use;

function get_mic_mapping(shot) {
	// data.config.micmapping_perframe is no longer used
	var mic_mapping = "12345";
	if (shot && shot.mic_mapping) mic_mapping = shot.mic_mapping;

	if (shot.source == "SOLO") {
		// #TE 2018-08-06: Quick and dirty workaround 
		//     to use SOLO and G2 at the same time on a S25...
//		mic_mapping = "58472613"; // DMC: I disagreee - can set "Wiring Harness" on each frame
	}
	return mic_mapping;
}
exports.get_mic_mapping = get_mic_mapping;

function get_mic_mapping_frame(frame) {
	// data.config.micmapping_perframe is no longer used
	var mic_mapping = "12345";
	if (frame && frame.geom && frame.geom.mic_mapping) mic_mapping = frame.geom.mic_mapping;
	
	return mic_mapping;
}
exports.get_mic_mapping_frame = get_mic_mapping_frame;

//
// return canonical mic positions
// shot.geom.
//   .sens_geom == "STD"
//
//   .sens_geom == "HBAR4","HBAR5","HBAR5","HBAR6"
//   .sens_geom == "HVBAR5", "HVBAR6"
//   .sens_geom == "CHAMBER"
//
function get_mic_pos(shot) {

	var geom = shot.geom;
	geom.sens_geom = geom.sens_geom; // make it exist, in case it is undefined
	var sens_geom = geom.sens_geom;
	
	var geom_is_chamber = "CHAMBER" == sens_geom;
	var geom_is_hbar = 0 == sens_geom.indexOf("HBAR");
	var geom_is_hvbar = 0 == sens_geom.indexOf("HVBAR");
	if (!geom_is_hbar && !geom_is_hvbar && !geom_is_chamber) sens_geom = "STD";
	var geom_is_std = "STD" == sens_geom;

	var mic_pos = [];

	if (geom_is_std) {
		// mic positions

		var spacing = shot.geom.chrono_spacing;
		var mic_w = shot.geom.mic_width;
		var mic_h = shot.geom.mic_height;

		// all standard physical positions
		//
		//    MIC2+6               MIC4+7
		//      .---------------------.
		//      |                     |
		//      |                     |
		//      |         **          |
		//      |        ****         |
		//      |         **          |
		//      |                     |
		//      |                     |
		//      .---------------------.
		//    MIC1+3               MIC5+8
		//
		mic_pos = [{x: -mic_w/2, y: -mic_h/2, z: 0}, // mic1 bottom left in-target-plane
			   {x: -mic_w/2, y: mic_h/2, z: 0}, // mic2 top left in-target-plane
			   {x: -mic_w/2, y: -mic_h/2, z: spacing},
			   {x: mic_w/2, y: mic_h/2, z: 0},
			   {x: mic_w/2, y: -mic_h/2, z: 0},
			   {x: -mic_w/2, y: mic_h/2, z: spacing},
			   {x: mic_w/2, y: mic_h/2, z: spacing},
			   {x: mic_w/2, y: -mic_h/2, z: spacing}];
	} else if (geom_is_hbar) {
		//
		// sensor bar underneath target has four chronos:
		//           .-------------.
		//           |             |
		//           |             |
		//           |     **      |            dist=Y0
		//           |     **      |              ^
		//           |     **      |              |
		//           |             |              |
		//           |             |              |
		//           .-------------.              |
		//                                        |
		//   MIC1+3   MIC2+6  MIC4+7   MIC5+8  ---.
		//
		//   x-coord= geom.(X13, X26, X47, X58)
		//   y-cord = -geom.Y0  ("Y0" is distance from sensor bar to nominal (0,0)
		//
		//   z-coord = 0 for back mics and geom.spacing for front mics
		var spacing = shot.geom.chrono_spacing;
		var y0 = shot.geom.Y0;
		var x13 = shot.geom.X13;
		var x26 = shot.geom.X26;
		var x47 = shot.geom.X47;
		var x58 = shot.geom.X58;
		mic_pos = [{x: x13, y: -y0, z: 0}, // mic1 leftmost in-target-plane
			   {x: x26, y: -y0, z: 0}, // mic2 next-leftmost in-target-plane
			   {x: x13, y: -y0, z: spacing},
			   {x: x47, y: -y0, z: 0},
			   {x: x58, y: -y0, z: 0},
			   {x: x26, y: -y0, z: spacing},
			   {x: x47, y: -y0, z: spacing},
			   {x: x58, y: -y0, z: spacing}];
	} else if (geom_is_hvbar) {
		// all standard physical positions
		//
		// sensor bar underneath target has four chronos:
		//           .-------------.
		//           |             |
		//           |             |
		//           |     **      |            dist=Y0
		//           |     **      |              ^
		//           |     **      |              |
		//           |             |              |
		//           |             |              |
		//           .-------------.              |
		//    MIC6 ^                  ^ MIC7      |
		//         |DY6           DY7 |           |
		//   MIC1+3   MIC2  MIC4    MIC5+8  ---.
		//
		//   x-coord= geom.(X13, X26, X47, X58)
		//   y-cord = -geom.Y0  ("Y0" is distance from sensor bar to nominal (0,0)
		//            +DY1 or DY2 etc....
		//
		//   z-coord = 0 for back mics and geom.spacing for front mics
		var spacing = shot.geom.chrono_spacing;
		var y0 = shot.geom.Y0;
		var x136 = shot.geom.X136;
		var x2 = shot.geom.X2;
		var x4 = shot.geom.X4;
		var x587 = shot.geom.X587;
		var dy6 = shot.geom.DY6 || 0;
		var dy7 = shot.geom.DY7 || 0;
		mic_pos = [{x: x136, y: -y0, z: 0}, // mic1 leftmost in-target-plane
			   {x: x2, y: -y0, z: 0}, // mic2 next-leftmost in-target-plane
			   {x: x136, y: -y0, z: spacing},
			   {x: x4, y: -y0, z: 0},
			   {x: x587, y: -y0, z: 0},
			   {x: x136, y: -y0+dy6, z: 0},
			   {x: x587, y: -y0+dy7, z: 0},
			   {x: x587, y: -y0, z: spacing}];
//			   console.log("DMC: HVBAR mic_pos=",mic_pos)
	} else { // "CHAMBER"
	}

	var i; for(i = 0; i < mic_pos.length; i++) if (mic_pos[i] == -1) mic_pos[i] = undefined;
	return mic_pos;
}
exports.get_mic_pos = get_mic_pos;

