/*******************************************************************************
 client_face.js: client class for displaying target faces

 $Author: dmc $
 $Revision: 3105 $
 $Date: 2022-09-09 09:25:20 -0300 (Fri, 09 Sep 2022) $
 $HeadURL: http://support2.silvermountaintargets.com/svn/smt/smtv2-lodi/public/javascripts/client_face.js $
*******************************************************************************/

// Face: a target board (geometry, common display, distance)

// everything in 'data' is synced with the server
// everything else is client-side only

/* target image mode
0: 1024 fixed size target, stretch image on short distance targets
1: 100 px/moa fixed resolution, large image on short distance targets
2: 1024 fixed size 10 moa center patch
*/

var face_mode = 0;


// none of them work quite right on Wayne's Kindle:
//
// 2-> square
// 1-> square
// 0-> 200 OK, 600 OK except for gray bounding square, 1000 y square
//

var g_shot_delay_remaining = 0; // global var, when nonzero indicates an F-Class 7-second delay is in effect


function ClientFace(data) {
	this.data = data;
	
	//log("creating face: " + data.name);

	this.crud = function(delta) {
		var msg = new_crud();
		msg.faces[this.id()] = delta;
		log("face.crud()", msg);
		g_socket.emit("crud", msg);
	}

	this.id = function() {
		return this.data._id;
	}

	this.mm_to_moa = 1 / (0.29088822 * data.distance);
	this.calc_mm_to_moa = function(dist_metres) {
		return 1 / (0.29088822 * dist_metres);
	};
	
	this.paper = null; // canvas for caching static graphics
	this.paper_image = null; // image representation of the canvas
	this.thumbnail = null; // canvas for caching small targetrow display
	this.thumbnail_image = null; // image of the canvas

	// f-class versions
	this.paper_f = null;
	this.paper_image_f = null;
	this.thumbnail_f = null;
	this.thumbnail_image_f = null;
	
	this.update_timer = null; // timer for updating automatically
	
	this.init_paper = function() {

		var board_w = this.data.targetboard.width;
		var board_h = this.data.targetboard.height;

		var args = {};
		
		// fixed size image for entire board
		if (face_mode == 0) {
			args.mm_to_px = 1024 / max(board_w, board_h);
			args.moa_to_px = args.mm_to_px / this.mm_to_moa;
		} else if (face_mode == 1 || face_mode == 2) {
			args.moa_to_px = 102.4;
			args.mm_to_px = this.mm_to_moa * args.moa_to_px;
		}
		
		if (face_mode == 0 || face_mode == 1) {
			args.w = round(board_w * args.mm_to_px);
			args.h = round(board_h * args.mm_to_px);
		} else if (face_mode == 2) {
			args.w = 1024;
			args.h = 1024;
		}
		
		args.cx = args.w/2;
		args.cy = args.h/2;

		//log("face paper size: " + args.w + " x " + args.h);

		this.paper = $("<canvas>").attr({width: args.w, height: args.h});
		
		var g = this.paper[0].getContext("2d");

		this.draw_target_board(g, args);
		this.draw_target_image(g, args);
		this.draw_target_zones(g, args, false);
		
		// convert canvas to image
		var img = this.paper[0].toDataURL("image/png");
		this.paper_image = new Image();
		this.paper_image.onload = function(event) {
			maintarget.update(); // force update
		}
		this.paper_image.onerror = function(event) {
			log("main canvas paper load error");
			log(event);
		}
		this.paper_image.src = img;

		// do this all over for f-class
		if (this.data.scoringzones_f) {
			
			this.paper_f = $("<canvas>").attr({width: args.w, height: args.h});
			
			var g = this.paper_f[0].getContext("2d");
	
			this.draw_target_board(g, args);
			this.draw_target_image(g, args);
			this.draw_target_zones(g, args, true);
			
			// convert canvas to image
			var img = this.paper_f[0].toDataURL("image/png");
			this.paper_image_f = new Image();
			this.paper_image_f.onload = function(event) {
				maintarget.update(); // force update
			}
			this.paper_image_f.onerror = function(event) {
				log("main canvas paper load error (f-class)");
				log(event);
			}
			this.paper_image_f.src = img;
			
		}

	}
	
	this.draw_target_paper = function(g, args, fclass) {
		
		// initialize if needed
		if (!ok(this.paper)) this.init_paper();
		
		// which paper image?
		var paper_to_use = fclass ? this.paper_image_f : this.paper_image;
		
		// image not ready yet, just wait
		if (!paper_to_use.naturalWidth) return;

		var board = this.data.targetboard;

		var rect;

		var centre = this.data.centrecoords || {x:0, y:0 };
		if (face_mode == 0 || face_mode == 1) {
			rect = {
				x: args.cx + args.mm_to_px * (-board.width/2 + centre.x),
				y: args.cy + args.mm_to_px * (-board.height/2 - centre.y),
				w: args.mm_to_px * board.width,
				h: args.mm_to_px * board.height
			};
		} else if (face_mode == 2) {
			rect = {
				x: args.cx - args.moa_to_px * 5,
				y: args.cy - args.moa_to_px * 5,
				w: args.moa_to_px * 10,
				h: args.moa_to_px * 10
			};
		}
		
		g.drawImage(paper_to_use, rect.x, rect.y, rect.w, rect.h);

	}
	
	this.init_thumbnail = function() {

		var board_w = this.data.targetboard.width;
		var board_h = this.data.targetboard.height;

		var args = {};

		// fixed size image for entire board
		args.mm_to_px = 512 / max(board_w, board_h);
		args.moa_to_px = args.mm_to_px / this.mm_to_moa;
		args.w = round(board_w * args.mm_to_px);
		args.h = round(board_h * args.mm_to_px);
		args.cx = args.w/2;
		args.cy = args.h/2;

		log("thumbnail size: " + args.w + " x " + args.h);
		
		this.thumbnail = $("<canvas>").attr({width: args.w, height: args.h});

		var g = this.thumbnail[0].getContext("2d");

		this.draw_target_board(g, args);
		this.draw_target_image(g, args);
		this.draw_target_zones(g, args, false);

		// convert canvas to image
		// dont force a reload - we don't want 10 reloads for no reason
		var img = this.thumbnail[0].toDataURL("image/png");
		this.thumbnail_image = new Image();
		this.thumbnail_image.src = img;
		
		// do over for f-class
		if (this.data.scoringzones_f) {

			this.thumbnail_f = $("<canvas>").attr({width: args.w, height: args.h});
	
			var g = this.thumbnail_f[0].getContext("2d");
	
			this.draw_target_board(g, args);
			this.draw_target_image(g, args);
			this.draw_target_zones(g, args, true);
	
			// convert canvas to image
			// dont force a reload - we don't want 10 reloads for no reason
			var img = this.thumbnail_f[0].toDataURL("image/png");
			this.thumbnail_image_f = new Image();
			this.thumbnail_image_f.src = img;
			
		}

	}

	this.draw_target_thumbnail = function(g, args, fclass) {

		// initialize as needed
		if (!ok(this.thumbnail)) this.init_thumbnail();
		
		fclass = fclass && !!this.data.scoringzones_f;
		var thumb_to_use = fclass ? this.thumbnail_f : this.thumbnail;
		var thumb_image_to_use = fclass ? this.thumbnail_image_f : this.thumbnail_image;
		
		var board = this.data.targetboard;
		var centre = this.data.centrecoords || {x:0, y:0 };

		var rect = {
			x: args.cx + args.mm_to_px * (-board.width/2 + centre.x),
			y: args.cy + args.mm_to_px * (-board.height/2 - centre.y),
			w: args.mm_to_px * board.width,
			h: args.mm_to_px * board.height
		};

		// draw thumbnail image if loaded
		if (thumb_image_to_use.naturalWidth)
			g.drawImage(thumb_image_to_use, rect.x, rect.y, rect.w, rect.h);
		else // otherwise draw the canvas itself
			g.drawImage(thumb_to_use[0], rect.x, rect.y, rect.w, rect.h);

	}
	
	this.draw_target_board = function(g, args) {

		var board = this.data.targetboard;
		var centre = this.data.centrecoords || {x:0, y:0 };

		g.beginPath();

		var rect = {
			x: args.cx + args.mm_to_px * (-board.width/2 + centre.x),
			y: args.cy + args.mm_to_px * (-board.height/2 - centre.y),
			w: args.mm_to_px * board.width,
			h: args.mm_to_px * board.height
		};

		g.rect(rect.x, rect.y, rect.w, rect.h);

		if (board.colour) {
			g.fillStyle = board.colour;
			g.fill();
		}
		if (board.bordercolour) {
			g.strokeStyle = board.bordercolour;
			g.lineWidth = (board.borderwidth * args.mm_to_px) || (args.moa_to_px * 0.2); // DMC: TODO
			g.stroke();
		}

	};

	this.draw_target_image = function(g, args) {
		if (this.data.image) {

			var centre = this.data.image.centre || {x:0, y:0 };
			var rect = {
				x: args.cx + args.mm_to_px * (-this.data.image.width/2 + centre.x),
				y: args.cy + args.mm_to_px * (-this.data.image.height/2 + centre.y),
				w: args.mm_to_px * this.data.image.width,
				h: args.mm_to_px * this.data.image.height
			};

			var src = this.data.image.src;
			var img = loaded_images[this.data.image.src];
			
			// check if need to request image from server
			if (!img) {
				log("requesting image " + src);
				img = new Image();
				loaded_images[src] = img;

				var _this = this;
				img.onload = function(event) { // when loaded...
					_this.paper = null; // invalidate so it is reinitialized
					_this.paper_f = null;
					_this.thumbnail = null;
					_this.thumbnail_f = null;
					maintarget.update(); // force update
				}
				img.onerror = function(event) { // invalid img will remain so there are no repeated requests
					log("img load error");
					log(event);
				}
				img.src = src; // start the request

			} else if (img.naturalWidth) { // image is valid
				g.drawImage(img, rect.x, rect.y, rect.w, rect.h);
			}
		}
	}

	this.draw_target_zones = function(g, args, fclass) {
		
		var zones = fclass ? this.data.scoringzones_f : this.data.scoringzones;
		
		for (var i = zones.length - 1; i >= 0; i--) {
			var z = zones[i];
			this.draw_target_zone(g, args, z);
		}
		
	};

	this.draw_target_zone = function(g, args, z) {

		var linewidth = ((z.borderwidth * args.mm_to_px) || (args.moa_to_px * 0.03));
		if (linewidth < 2) linewidth = 2;

		if (z.colour) {
			g.fillStyle = z.colour;
//			g.fill();
		}

		if (z.bordercolour) {
			g.strokeStyle = z.bordercolour;
			g.lineWidth = linewidth;
//			g.stroke();
		}

		if (z.type.name == "ring") {
			g.beginPath();

			var center = z.centre ? z.centre : {x:0, y:0};
			var radius = z.type.radius ? z.type.radius : (z.type.dia/2);
			radius -= (linewidth/args.mm_to_px) / 2;
			if (radius < 0) radius = 0;

			var arc = {
				x: args.cx + args.mm_to_px * center.x,
				y: args.cy - args.mm_to_px * center.y, // DMC: tricksy: "-y"
				r: args.mm_to_px * radius
			};

			g.arc(arc.x, arc.y, arc.r, 0, 2*Math.PI);

		} else if (z.type.name == "rect") {
			g.beginPath();

			var center = z.centre ? z.centre : {x:0, y:0};

			var rect = {
				x: args.cx + args.mm_to_px * (-z.type.width/2 + center.x),
				y: args.cy + args.mm_to_px * (-z.type.height/2 - center.y), // DMC: tricksy: "-y"
				w: args.mm_to_px * z.type.width,
				h: args.mm_to_px * z.type.height
			};

			g.rect(rect.x, rect.y, rect.w, rect.h);

		} else if (z.type.name == "poly") {
			g.beginPath();

			var center = {x: args.cx, y: args.cy};
			var offset = z.type.shiftamt ? {x: z.type.shiftamt[0] * args.mm_to_px,
				y: -z.type.shiftamt[1] * args.mm_to_px} : {x:0, y:0}; // DMC 3-June-2015 very tricksy stuff here
			// ... the y-axis runs *BACKWARDS*, since it is in native pixel coords.

			g.moveTo(center.x + offset.x + args.mm_to_px * z.type.vertices[0][0],
					 center.y + offset.y - args.mm_to_px * z.type.vertices[0][1]);

			for (var j = 1; j < z.type.vertices.length; j++) {
				g.lineTo(center.x + offset.x + args.mm_to_px * z.type.vertices[j][0],
						 center.y + offset.y - args.mm_to_px * z.type.vertices[j][1]);
			}

		} else if (z.type.name == "and" || z.type.name == "or") {

			for (var i = z.type.members.length - 1; i >= 0; i--) {
				var zz = z.type.members[i];
				this.draw_target_zone(g,args,zz);
			}
			return; // no actual drawing done, skip the closePath() etc below.

		} else if (z.type.name == "text") {
			g.beginPath();
			var center = z.centre ? z.centre : {x:0, y:0};
			var canv_x = args.cx + args.mm_to_px * center.x;
			var canv_y = args.cy - args.mm_to_px * center.y; // DMC: tricksy: "-y"
			var canv_sz = args.mm_to_px * (z.type.text_size || 10);
			g.font = canv_sz+'px ' + (z.type.text_font || 'Helvetica');
			g.textAlign = z.type.textAlign || "center";
			// default: fillText, but can turn off
			z.type.fillText != 'no' && g.fillText(z.type.text,canv_x,canv_y);
			// default: no strokeText, but can turn on
			z.type.strokeText && g.strokeText(z.type.text,canv_x,canv_y);




		} else { // unrecognized zone type
//console.log("dmc-draw_target_zone",z.type.name);
			return;
		}

		g.closePath();

		if (z.colour) {
//			g.fillStyle = z.colour;
			g.fill();
		}

		if (z.bordercolour) {
//			g.strokeStyle = z.bordercolour;
//			g.lineWidth = linewidth;
			g.stroke();
		}

	};
	
	this.draw_target_grid = function(g, args) {
	
	
		// if (args.face.data.grid && args.face.data.grid.disabled) return; // targetface says no-grid
		if (g_config.option_grid_unit == 'none') return; // ADMIN Officer says no-grid
		
		g.beginPath();
		
		var board = {
			x: args.cx + args.mm_to_px * (-this.data.targetboard.width/2 + this.data.centrecoords.x),
			y: args.cy + args.mm_to_px * (-this.data.targetboard.height/2 - this.data.centrecoords.y),
			w: args.mm_to_px * this.data.targetboard.width,
			h: args.mm_to_px * this.data.targetboard.height,
		};
		

		// grid resolution (default 1 moa)
		var res = {x: 1, y: 1};
		if (g_config.option_grid_unit == "tenthmil") res.x = res.y = mil_to_moa(.1); // tenth-mil
		if (g_config.option_grid_unit == "RMR") res.x = res.y = mil_to_moa(.3); // three tenths-mil
		if (g_config.option_grid_unit == "mil") res.x = res.y = mil_to_moa(1); // full mil
		// if (g_config.option_grid_unit == "mil20") res.x = res.y = mil_to_moa(.05); // twentieth-mil
		// if (g_config.option_grid_unit == '1"') res.x = res.y = 1 *1/(args.face.data.distance*3.280839/3/100*1.047);
		// if (g_config.option_grid_unit == '6"') res.x = res.y = 6 *1/(args.face.data.distance*3.280839/3/100*1.047);

		// face options can override (for now...)
		if (this.data.options) {
			if (this.data.options.grid_resolution_moa) {
				res.x = res.y = this.data.options.grid_resolution_moa;
			}
			if (this.data.options.grid_resolution_moa_x) {
				res.x = this.data.options.grid_resolution_moa_x;
			}
			if (this.data.options.grid_resolution_moa_y) {
				res.y = this.data.options.grid_resolution_moa_y;
			}
		}

		// grid lines spaced every (res_x,res_y) MOA
		var grid_x_range = Math.floor((board.w/2) / (args.moa_to_px*res.x));
		var grid_y_range = Math.floor((board.h/2) / (args.moa_to_px*res.y));
		var n_gridlines_left = (this.data.options && this.data.options.n_gridlines_left) || grid_x_range;
		var n_gridlines_right = (this.data.options && this.data.options.n_gridlines_right) || grid_x_range;
		var n_gridlines_up = (this.data.options && this.data.options.n_gridlines_up) || grid_y_range;
		var n_gridlines_down = (this.data.options && this.data.options.n_gridlines_down) || grid_y_range;

		for (var moa_x = -n_gridlines_left*res.x; moa_x <= (n_gridlines_right+0.49)*res.x; moa_x += res.x) {
			var x = round(board.x + board.w/2 + moa_x * args.moa_to_px);
			if (x < 0 || x > args.w) continue;
			var down_y = round(board.y + board.h/2 + n_gridlines_down*res.y * args.moa_to_px);
			var up_y = round(board.y + board.h/2 + -n_gridlines_up*res.y * args.moa_to_px);
			g.moveTo(x, down_y);
			g.lineTo(x, up_y);
		}
	
		for (var moa_y = n_gridlines_down*res.y; moa_y >= -(n_gridlines_up+0.49)*res.y; moa_y -= res.y) {
			var y = round(board.y + board.h/2 + moa_y * args.moa_to_px);
			if (y < 0 || y > args.h) continue;
			var left_x = round(board.x + board.w/2 + -n_gridlines_left*res.x * args.moa_to_px);
			var right_x = round(board.x + board.w/2 + n_gridlines_right*res.x * args.moa_to_px);
			g.moveTo(left_x, y);
			g.lineTo(right_x, y);
		}
	
		g.strokeStyle = "#222";
		g.lineWidth = 3;
		g.stroke();
		g.strokeStyle = "#DDD";
		g.lineWidth = 1.5;
		g.stroke();
		
	};
	
	this.draw_target_shots = function(g, args, fclass) {
		
		// shot size = fixed bullet diameter in mm
		var bullet_diameter = 7.82;
		var default_shot_scale = bullet_diameter * args.face.mm_to_moa;
		
		// set minimum size, match parameter in maintarget
		var shot_scale_min = maintarget.zoom / 20;
		if (default_shot_scale < shot_scale_min) default_shot_scale = shot_scale_min;
		
		default_shot_scale *= 4; // needs to match existing display ratios

		var shots = args.string.shots();
		
		var update_again_soon = false;
		
		var shot_delay_remaining = 0;
		var shot_delay_needed = false;
		
		var now = (new Date()).getTime() - cur_ts_local_error;
		
		// display shot dot
		for (var i in shots) {
			var shot = shots[i];
			if (!shot) continue;
			
			// individually sized shots
			var shot_scale = default_shot_scale;
			if (shot.bullet_dia) shot_scale *= shot.bullet_dia / 7.82;
			
			var px = args.cx + args.mm_to_px * shot.x;
			var py = args.cy - args.mm_to_px * shot.y;
			
			// dont show invalid shots to a user

			if (!display_admin_mode() && !display_info_mode() && !shot.valid) {
				continue; // only admin/info can see invalid shots
			}

			if (!shot.v) continue; // must have an accepted velocity to display
			if (!shot.show_shotdisc) continue;
//			if (shot.strikeout) continue;

			// do not show new shots for 7 seconds
			if (isShotHiddenByDelay(shot, args.string)) {
				var time_remaining = shot.ts + g_config.shot_display_delay * 1000 - now;
				if (time_remaining > 0) shot_delay_needed = true;
				if (g_shot_delay_remaining == 0 && shot_delay_needed) g_shot_delay_remaining = time_remaining;
//shot_delay_remaining = time_remaining;
				if (time_remaining > shot_delay_remaining) shot_delay_remaining = time_remaining;
 
				update_again_soon = true;
				continue;
			}
			
			var recent_shot_age = 1000* (g_config.shot_display_delay > 0 ? 2 * g_config.shot_display_delay : 7)
			var shot_is_recent = live_mode() && now < shot.ts + recent_shot_age && now > shot.ts - 5000
			var shot_is_selected = (shot == args.string.cur_shot());

			// draw shot highlights
			if (!printing && (shot_is_selected || shot_is_recent)) {

				// highlighting recent shots require another update later
				if (shot_is_recent) update_again_soon = true;

				// selected shot is off screen: draw arrow
				if (shot_is_selected && (px < -25 || px > args.w + 25 || py < -25 || py > args.h + 25)) {

					var dx = (px < -25)?(-1):((px > args.w + 25)?(1):(0));
					var dy = (py < -25)?(-1):((py > args.h + 25)?(1):(0));

					var ex = (px < -25)?(0):((px > args.w + 25)?(args.w):(px));
					var ey = (py < -25)?(0):((py > args.h + 25)?(args.h):(py));

					var theta = 20 * 3.1415 / 180;
					var p1x = dx*cos(theta) - dy*sin(theta);
					var p1y = dx*sin(theta) + dy*cos(theta);
					var p2x = dx*cos(-theta) - dy*sin(-theta);
					var p2y = dx*sin(-theta) + dy*cos(-theta);

					g.beginPath();
					var length = args.moa_to_px * 0.75 * shot_scale;
					if (length > 100) length = 100;
					g.moveTo(ex - p1x * length, ey - p1y * length); g.lineTo(ex, ey);
					g.moveTo(ex - p2x * length, ey - p2y * length); g.lineTo(ex, ey);
					g.moveTo(ex - dx * length * 3, ey - dy * length * 3); g.lineTo(ex, ey);
					g.strokeStyle = "#F44";
					g.lineWidth = args.moa_to_px * 0.05 * shot_scale;
					g.stroke();

				}

				// draw red cross -- only for selected shot
				if (shot_is_selected) {
					g.beginPath();
					var cross = args.moa_to_px * 0.6 * shot_scale;
					if (!shot_is_recent) cross *= 0.4;
					g.moveTo(px - cross, py - cross); g.lineTo(px + cross, py + cross);
					g.moveTo(px + cross, py - cross); g.lineTo(px - cross, py + cross);
					g.strokeStyle = "#F44";
					g.lineWidth = args.moa_to_px * 0.08 * shot_scale;
					g.stroke();
				}

			}

			// nothing else for other shooter's shots
			if (!display_admin_mode()) {
				if (cur_string().data.shooter_count > 1 && shot.shooter_num != cur_shooter) continue;
			}

			// draw path to measured location
			if (display_admin_mode(true) && shot.measured) {
				
				var mx = args.cx + args.mm_to_px * shot.measured_x;
				var my = args.cy - args.mm_to_px * shot.measured_y;
				
				// draw line to measured position
				g.beginPath();
				g.moveTo(px, py);
				g.lineTo(mx, my);
				g.strokeStyle = "#FF0";
				g.lineWidth = 5;
				g.stroke();
/*
				// draw shot dot
				g.beginPath();
				g.arc(mx, my, args.moa_to_px * 0.06, 0, 2*Math.PI);
				g.fillStyle = "#FF0";
				g.fill();
*/
			}			

			// plot adjustment
			if (plotting.enable_view) {
				
				var adj_x = 0;
				var adj_y = 0;
				
				var params = args.string.data.shot_params;

				// plot scope setting
				if (plotting.args_x.pan_scope) {
					adj_x += plotting.args_x.pan_scope_x;
				} else if (params.scope) {
					if (typeof(params.scope_x) == "object") {
						adj_x += params.scope_x[cur_shooter];
					} else {
						adj_x += params.scope_x;
					}
				}
				
				if (plotting.args_y.pan_scope) {
					adj_y += plotting.args_y.pan_scope_y;
				} else if (params.scope) {
					if (typeof(params.scope_y) == "object") {
						adj_y += params.scope_y[cur_shooter];
					} else {
						adj_y += params.scope_y;
					}
				}

				// plot hold setting
				if (maintarget.args.pan_hold) {
					adj_x += maintarget.args.pan_hold_x;
					adj_y += maintarget.args.pan_hold_y;
				} else if (params.hold) {
					adj_x += params.hold_x;
					adj_y += params.hold_y;
				}
				
				if (shot.scope) {
					if (typeof(shot.scope_x) == "object") {
						adj_x -= shot.scope_x[cur_shooter];
						adj_y -= shot.scope_y[cur_shooter];
					} else {
						adj_x -= shot.scope_x;
						adj_y -= shot.scope_y;
					}
				}
				if (shot.hold) {
					adj_x -= shot.hold_x;
					adj_y -= shot.hold_y;
				}

				px += args.moa_to_px * adj_x;
				py -= args.moa_to_px * adj_y;
			}
			
			var drag_cal_this_shot = (maintarget.args.drag_cal_active && shot == args.string.data.shots[maintarget.args.drag_cal_shot]);
			
			// drag calibration adjustment
			if (drag_cal_this_shot) {
				px += args.mm_to_px * maintarget.args.drag_cal_x;
				py -= args.mm_to_px * maintarget.args.drag_cal_y;
			}

			var shotcolour = shot.colour;
			if (shotcolour == "default") shotcolour = undefined;
			// measured
			if (drag_cal_this_shot) {
				g.beginPath();
				var radius = args.moa_to_px * 0.18 * shot_scale;
				g.arc(px, py, radius, 0, 2*Math.PI);
				g.fillStyle = "#4F4";
				g.fill();
			} else if (display_admin_mode(true) && shot.measured) { // yellow ring around cal'd shots
				g.beginPath();
				var radius = args.moa_to_px * 0.18 * shot_scale;
				g.arc(px, py, radius, 0, 2*Math.PI);
				g.fillStyle = "#FF4";
				g.fill();
				//DMC start 6-Oct-2018
			} else if ((display_info_mode()||display_admin_mode(true)) && 
			  (shot.invalid_reasons && shot.invalid_reasons.length && shot.invalid_reasons.indexOf("V_ERR_TOOBIG") != -1)
			) { //:
				g.beginPath();
				var radius = args.moa_to_px * 0.22 * shot_scale;
				g.arc(px, py, radius, 0, 2*Math.PI);
				g.fillStyle = "orange";
				g.fill();
				// DMC end 6-Oct-2018
			} else if (display_info_mode() && shot.screened_out_theta) { // purple ring around crossfire shots:
				g.beginPath();
				var radius = args.moa_to_px * 0.22 * shot_scale;
				g.arc(px, py, radius, 0, 2*Math.PI);
				g.fillStyle = "purple";
				g.fill();
			} else if (!printing && shot == args.string.cur_shot()) { // selected
				g.beginPath();
				var radius = args.moa_to_px * 0.18 * shot_scale;
				g.arc(px, py, radius, 0, 2*Math.PI);
				g.fillStyle = "#F44";
				g.fill();
			}

			g.beginPath();
			g.arc(px, py, args.moa_to_px * 0.12 * shot_scale, 0, 2*Math.PI);

			// draw shot dot and outline

			g.lineWidth = args.moa_to_px * 0.02 * shot_scale;
			if (plotting.enable_view) { // plot mode display
				g.fillStyle = "#444";
				g.fill();
				g.strokeStyle = "#88F";
				g.stroke();
			} else { // normal
				// recent shots: draw 'disc' in same red as cross
				if (shot_is_recent) {
					g.fillStyle = "#F44";
				} else {
					g.fillStyle = shotcolour || "#444";
				}
				g.fill();
				g.strokeStyle = "#BBB";
				g.stroke();
			}
			
			// draw shot text
			if (maintarget.zoom > maintarget.zoom_min) {
				g.fillStyle = "#FFF";
				g.textAlign = "center";
				g.textBaseline = "middle";
				g.font = Math.round(args.moa_to_px * 0.18 * shot_scale) + "px Arial";
				g.fillText(shot.text, px, py);
			}

		}
		
		// todo: this drawing should be outside of this function, but w/e for now

		// draw mic tap test graphic
		if (display_admin_mode(true) && live_mode() && args.string.data.taps !== undefined) { // shots exist
			var taps = args.string.data.taps;

			var tap_last = taps[taps.length - 1]; // most recent shot

			var now = (new Date()).getTime() - cur_ts_local_error;
			if (now < tap_last.ts + 5000 && now > tap_last.ts - 5000) { // happened within 5 seconds
			
				var mic_tap = get_mic_tap_num(tap_last.mics); // defined in utils.js
				
				// draw it
				if (mic_tap) {
					g.fillStyle = "#4F4";
					g.textAlign = "center";
					g.textBaseline = "middle";
					g.font = Math.round(args.moa_to_px * 4) + "px Arial";
					g.fillText(mic_tap, args.cx, args.cy);
					update_again_soon = true;
				}
				
			}
		}
		
		// draw connection error graphic
		if (!g_connected) {
			
			g.fillStyle = "#F44";
			g.textAlign = "center";
			g.textBaseline = "middle";
			g.font = round(args.h * 1.5) + "px Arial";
			g.fillText("X", args.cx, args.cy); // red-X
			
		}
		
		if (shot_delay_remaining) {
			g.textAlign = "center";
			g.textBaseline = "middle";
			
			var cx = maintarget.div.width() * 0.5;
			var cy = maintarget.div.height() * 0.65;
			var size = round(maintarget.div.height() * 0.08);
			var w = size * 6;
			var h = size * 2;

			g.font = size + "px Arial";
			var time_left = round(shot_delay_remaining * 0.001);
			if (time_left < 1) time_left = 1;
			var str = "Wait: " + round(shot_delay_remaining * 0.001) + " s"
			
			var cx = maintarget.div.width() * 0.5;
			var cy = maintarget.div.height() * 0.65;

			g.beginPath();
			g.fillStyle = "#444";
			g.rect(cx - w/2, cy - h/2, w, h);
			g.fill();

			g.fillStyle = "#FFF";
			g.fillText(str, cx, cy);
		}

		if (update_again_soon) {
			clearTimeout(this.update_timer);
			this.update_timer = setTimeout(function() {
				maintarget.update();
				targetrow.update();
			}, 500);
		}

	};
    
	this.draw_only_selected_shot = function(g, args, fclass) {
		
		// check all shots for those on this target
		// draw selected shot and return
		var shot = args.string.cur_shot();
        
		if (shot && shot.v) {

			if (shot.strikeout) return;
			
			// do not show new shots for 7 seconds
			// maintarget will handle autorefresh for us
			//var now = Date.now();
            //var now = (new Date()).getTime() - cur_ts_local_error;
            //fclass = draw_fclass_target_test(args.string);
        
			if (isShotHiddenByDelay(shot, args.string)) {
				return;
			}
			
			var px = args.cx + args.mm_to_px * shot.x;
			var py = args.cy - args.mm_to_px * shot.y;
	
			// draw cross behind selected shot
			g.beginPath();
	
			var cross = args.moa_to_px * 2.0;
			
			//g.moveTo(px - cross, py - cross); g.lineTo(px + cross, py + cross);
			//g.moveTo(px + cross, py - cross); g.lineTo(px - cross, py + cross);
			g.moveTo(px, py - cross); g.lineTo(px, py + cross);
			g.moveTo(px + cross, py); g.lineTo(px - cross, py);
	
			g.strokeStyle = "#F44";
			g.lineWidth = 3; // args.moa_to_px * 0.25;
			g.stroke();
	
			// draw special shot dot
			g.beginPath();
	
			var radius = 6; // args.moa_to_px * 0.5;
			g.arc(px, py, radius, 0, 2*Math.PI);
	
			g.fillStyle = "#F44";
			g.fill();

			return;
		}
	
	};

	// draw the special recent shot on a device thumbnail
	this.draw_device_recent_shot = function(g, args) {

		if (!args.device.data.shot_valid) return;
		
		var px = args.cx + args.mm_to_px * args.device.data.shot_x;
		var py = args.cy - args.mm_to_px * args.device.data.shot_y;

		// draw cross behind selected shot
		g.beginPath();

		var cross = args.moa_to_px * 1.0;
		
		g.moveTo(px - cross, py - cross); g.lineTo(px + cross, py + cross);
		g.moveTo(px + cross, py - cross); g.lineTo(px - cross, py + cross);

		g.strokeStyle = "#F44";
		g.lineWidth = args.moa_to_px * 0.1;
		g.stroke();

		// draw special shot dot
		g.beginPath();

		var radius = args.moa_to_px * 0.3;
		g.arc(px, py, radius, 0, 2*Math.PI);

		g.fillStyle = "#F44";
		g.fill();
		
	};
	
	this.draw_stats_circle = function(g, args) {

		if (!args.string.data.use_string_stats) return;
		if (!args.string.data.stats_circle_center) return;
		if (!g_config.option_shot_circle) return;

		var px = 0;
		var py = 0;
		var rx = 0;

        if (
            cur_shooter !== undefined 
            && args.string.data.shooterStats 
            && args.string.data.shooterStats[cur_shooter] !== undefined 
        ) {
            if (!args.string.data.shooterStats[cur_shooter].use_string_stats) return;
            var stats = args.string.data.shooterStats[cur_shooter];
            px = args.cx + args.mm_to_px * stats.stats_circle_center.x;
            py = args.cy - args.mm_to_px * stats.stats_circle_center.y;
            rx = args.mm_to_px * stats.stats_circle_radius;
        } else {
            px = args.cx + args.mm_to_px * args.string.data.stats_circle_center.x;
            py = args.cy - args.mm_to_px * args.string.data.stats_circle_center.y;
            rx = args.mm_to_px * args.string.data.stats_circle_radius;
        }
        
		
		g.beginPath();
		g.arc(px, py, rx, 0, 2*Math.PI);

		g.strokeStyle = "#FF8";
		g.lineWidth = args.moa_to_px * 0.05;
		g.stroke();
		
	}

	this.draw_stats_center = function(g, args) {

		if (!args.string.data.use_string_stats) return;
		if (!args.string.data.stats_group_center) return;
		if (!g_config.option_group_center) return;

		var px = 0;
		var py = 0;

        if (
            cur_shooter !== undefined 
            && args.string.data.shooterStats 
            && args.string.data.shooterStats[cur_shooter] !== undefined 
        ) {
            if (!args.string.data.shooterStats[cur_shooter].use_string_stats) return;
            var stats = args.string.data.shooterStats[cur_shooter];
            px = args.cx + args.mm_to_px * stats.stats_group_center.x;
            py = args.cy - args.mm_to_px * stats.stats_group_center.y;
        } else {
            px = args.cx + args.mm_to_px * args.string.data.stats_group_center.x;
            py = args.cy - args.mm_to_px * args.string.data.stats_group_center.y;
        }

		g.beginPath();

		var cross = args.moa_to_px * 1;
		g.moveTo(px, py - cross); g.lineTo(px, py + cross);
		g.moveTo(px - cross, py); g.lineTo(px + cross, py);

		g.strokeStyle = "#8F8";
		g.lineWidth = args.moa_to_px * 0.1;
		g.stroke();


	}
	
}

